/*  Authors:    WeiChung Hsu
                Tz-Yang Tang
                Bhashinee Garg
*/


{
    import java.io.*;
    import java.util.*;
}


//====================LEXER==========================================

class GECLLexer extends Lexer;
options 
{ 
    k = 2; 
    testLiterals = false;          // do not check matched tokens for (Parser) literals
    charVocabulary = '\3'..'\377'; // Using ASCII chars
	testLiterals = false;
}
    
// Identifier
ID  options {	testLiterals = true;    }
    :   ( 'a'..'z' | 'A'..'Z' | '_' )
        ( 'a'..'z' | 'A'..'Z' | '_' | '0'..'9' )*
    ;


// Comments
COMMENT
: "//" (~'\n')* '\n'
{ newline(); $setType(Token.SKIP); }
;


// String
// String constants must be contained on a single line and
// may contain double quotes, e.g.,
// "This is a constant with ""double quotes"""
StringConstant
: '"'! 
( ~('"' | '\n')
| ('"'! '"')
)* '"'!
;


// White space
WS
: ( ' '
| '\t'
| '\f'
// handle newlines
| ( "\r\n" // Evil DOS
| '\r' // Macintosh
| '\n' // Unix (the right way)
)
{ newline(); }
)
{ $setType(Token.SKIP); }
;

// Numbers               
// protected ==> protected lexer rules can be used by other lexer rules, but do not return tokens by themselves
protected DIGITS : ('0'..'9')+;

INTEGER 
    : DIGITS ('.' DIGITS {$setType(FLOAT);})? (UNIT {$setType(DIST);} )?;
    
// Symbols
ASSIGN 			: '=' 	; 
MOD             : '%' 	;
DEC             : "--" ;
MOD_ASSIGN      : "%=" 	;
LTE             : "<=" ;
PLUS            : '+' ;
LOR             : "||" ;
PLUS_ASSIGN     : "+=" ;
EQUAL           : "==" ;
NOT_EQUAL       : "!=";
PERIOD          : '.'	;
MINUS           : '-' ;
LAND            : "&&" ;
MINUS_ASSIGN    : "-=" ;
GT              : ">" ;
STAR            : '*' ;
LNOT            : '!' ;
STAR_ASSIGN     : "*=" ;

LT              : "<" ;
DIV             : '/' ;
INC             : "++" ;
DIV_ASSIGN      : "/=" ;
GTE             : ">=" ;

// Other symbols
SEMI: ';' ;
LPAREN : '(' ;
RPAREN : ')' ;
LCURLY : '{' ; 
RCURLY : '}' ; 
LBRACK : '[' ; 
RBRACK : ']' ; 
COMMA : ',' ;


protected
UNIT 
    :   "km"
    |   "ft"
    |	"nm"
    |	'm'
    |	"mi"        
    ;

//==================PARSER=========================

class GECLParser extends Parser;

options
    {
        k = 2;
        buildAST = true;
    }
tokens
    {
        SHAPE;
        NStatementExpr;
		NEmptyExpression;
		NStringSeq;
		NExpressionGroup;
		NFunctionCallArgs;
		NPostfixExpr;
		NUnaryExpr;
		NCommaExpr;
		PROGRAM_AST;
    }
	
program
    : 	stmts
        { #program = #([PROGRAM_AST], program); }
    ;
	
stmts
	:   (statement (WS)*)+ EOF!
	;
	

//stmt
statement
    :   SEMI                    // Empty statements
	|	distance SEMI!
    |   direction SEMI!
    |   coordinates SEMI!
    |   points SEMI!
	|	lines SEMI!
	|	overlays SEMI!
	|	perspective SEMI!
	|	places SEMI!
	|	structure SEMI!
    |   printout SEMI!
    |   assigncity SEMI!
	|   changeobj SEMI!   
	|	declarations SEMI!
	|	read3dfile SEMI!
	|   expr SEMI!               { ## = #( #[NStatementExpr], ## ); } // Expressions
    |   "while"^ LPAREN! expr RPAREN! block_statement
	|   "do"^ block_statement "while" LPAREN! expr RPAREN! SEMI!
	|   "continue" SEMI!
    |   "break" SEMI!
    |   "return"^ ( expr )? SEMI!
	|   "if"^ LPAREN! expr RPAREN! block_statement  
                ( //standard if-else ambiguity
                  options { warnWhenFollowAmbig = false; } :
                "else" block_statement )?
    ;
		
	
block_statement
			:	statement
			|	LCURLY	( statement )+ RCURLY
			;
			
declarations
	: "int"^ ID ASSIGN! INTEGER
	| "float"^ ID ASSIGN! FLOAT
	;
read3dfile
	: "readfile"^ ID (FLOAT | ID) (FLOAT | ID) (FLOAT | ID) (FLOAT | ID) StringConstant
	;

distance
    :   "dist"^ ID ("IS"! DIST)?
    ;

direction
    :   "dir"^ ID "IS"! dir_spec
    ;
  
dir_spec    
    :   INTEGER
    |   FLOAT
    |   compass_dir
    ;


compass_dir 
    :   "N"
    |   "S"
    |   "W"
    |   "E"
    |   "NE"
    |   "NW"
    |   "SE"
    |   "SW"
    ;

coordinates
    :   "coord"^ ID ("IS"! coord_params)?
    ;

coord_params 
    :   (FLOAT | (MINUS FLOAT)) COMMA! (FLOAT | (MINUS FLOAT)) (COMMA! DIST)?
    ;

points
    :	"point"^ (ID "IS"! ID)
	;

lines
	:	"line"^ ID "IS"! (ID)+ 
	;

overlays
	:	"overlay"^ ID "IS"! ID ID ID ID COMMA! StringConstant COMMA! dir_spec
	;

perspective
	:	"persp"^ ID "IS"! ID COMMA! DIST COMMA! dir_spec COMMA! dir_spec
	;

structure 
	:	"struct"^ ID ("IS"! shape COMMA! ID)?
	;

shape
	:	"square" FLOAT COMMA! FLOAT
	|	"rectangle" FLOAT COMMA! FLOAT COMMA! FLOAT
	|	"triangle" FLOAT COMMA! FLOAT
	|	"octagon" FLOAT COMMA! FLOAT
         { #shape = #([SHAPE], shape); }
	;

places
	:	"place"^ ID "IS"! ID COMMA! StringConstant ( COMMA! ID )?
	;

printout
	: "printk"^ ID 
	;

assigncity
	: "AsigCity"^ ID "IS"! LPAREN! StringConstant COMMA! StringConstant RPAREN!
	;

changeobj
    :   ID "IS"^ ID ((STAR INTEGER) | (PLUS ((DIST)|(LPAREN! (ID |DIST ) COMMA! (ID |dir_spec ) RPAREN!))) )
    ;
	
	
expr
    :   assignExpr (options { warnWhenFollowAmbig = false; } :
                            c:COMMA^ { #c.setType(NCommaExpr); } assignExpr         
                       )*
    ;
		
assignExpr
    :    conditionalExpr ( a:assignOperator! assignExpr { ## = #( #a, ## );} )?
    ;
	
assignOperator
    :    ASSIGN
    |    DIV_ASSIGN
    |    PLUS_ASSIGN
    |    MINUS_ASSIGN
    |    STAR_ASSIGN
    |    MOD_ASSIGN
    ;

conditionalExpr
    :       logicalOrExpr
            ( QUESTION^ expr COLON! conditionalExpr )?
    ;		
constExpr
    :    conditionalExpr
    ;
	
logicalOrExpr
    :   logicalAndExpr ( LOR^ logicalAndExpr )*
    ;

logicalAndExpr
    :   equalityExpr ( LAND^ equalityExpr )*
    ;

equalityExpr
    :   relationalExpr ( ( EQUAL^ | NOT_EQUAL^ ) relationalExpr )*
    ;	

relationalExpr
    :   additiveExpr ( ( LT^ | LTE^ | GT^ | GTE^ ) additiveExpr )*
    ;

additiveExpr
    :   multExpr ( ( PLUS^ | MINUS^ ) multExpr )*
    ;

multExpr
    :   unaryExpr ( ( STAR^ | DIV^ | MOD^ ) unaryExpr )*
    ;	

unaryExpr
    :   postfixExpr
    |   INC^ unaryExpr
    |   DEC^ unaryExpr
    |   u:unaryOperator castExpr { ## = #( #[NUnaryExpr], ## ); }
    ;

castExpr
    :   unaryExpr
    ;

unaryOperator
        :       STAR
        |       PLUS
        |       MINUS
        |       LNOT
        ;
			
postfixExpr
    :   primaryExpr ( postfixSuffix {## = #( #[NPostfixExpr], ## );} )?
    ;
	
postfixSuffix
    :   (LBRACKET expr RBRACKET | INC | DEC )+
    ;		

primaryExpr
    :   ID
    |   charConst
    |   INTEGER
    |   FLOAT
    |   StringConstant
    |   LPAREN! expr RPAREN!    { ## = #( #[NExpressionGroup, "("], ## ); }
    ;
			
argExprList
    :   ID ( COMMA! ID )*
    ;

protected
charConst
    :   CharLiteral
    ;
stringConst
    :   (StringLiteral)+         { ## = #(#[NStringSeq], ##); }
    ;
    
   
//==================TREE WALKER WITH STATIC SEMANTIC CHECKER=========================

{
	import java.io.PrintWriter;
	import java.io.IOException;
	import java.lang.Object;
	import wdcor.WorldCoords;
}

class GECLWalker extends TreeParser;
{
    java.util.Hashtable dict = new java.util.Hashtable();
    java.util.Hashtable PerspTab = new java.util.Hashtable();
    int count=0;
    int countp=0;
    GECLdatatypes[] b = new GECLdatatypes[100];
    GECLdatatypes[] tm = new GECLdatatypes[100];
    PerspectType[] pr = new PerspectType[100];
    Distance sep_dist;
    GECLlibrary kk = new GECLlibrary();
	
	
	int tabs = 0;		
	int if_break = 0;
	int if_continue = 0;
	int if_skip_rcurly = 0;
	int if_state = 1;
	
	float recursive_f(AST rec_root)
	{
		AST e1, e2;
		int entry_index;
        
		//first check if this root is binary operator, if yes, then it will recursively go to next level
		//if not, then just evaluate and return the value
		if(rec_root == null)
			return 0;
		else if(rec_root.getText().length()==0)
		{
			return recursive_f(rec_root.getFirstChild());
		}
		else if(rec_root.getText().equals("-"))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			return (recursive_f(e1) - recursive_f(e2));
		}
		else if(rec_root.getText().equals("*"))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			return (recursive_f(e1) * recursive_f(e2));
		}
		else if(rec_root.getText().equals("+"))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			return (recursive_f(e1) + recursive_f(e2));
		}
		else if(rec_root.getText().equals("/"))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			return (recursive_f(e1) / recursive_f(e2));
		}	
		else if(rec_root.getText().equals("="))
		{
			// assign to left node, left node must be ID
			e1 = (AST) rec_root.getFirstChild();	
			e2 = (AST) e1.getNextSibling();
			
			if( !(dict.containsKey(e1.getText()))) //this is number, make no sense to assign to number, so skip assing operator
			{			
				recursive_f(e2);
				return Float.parseFloat(e1.getText());
			}
			else // this is ID, could be assigned value
			{
				entry_index=(Integer)dict.get( e1.getText());				
				if(b[entry_index].type == "int")
				{
					b[entry_index].int_value = (int) recursive_f(e2);
					return b[entry_index].int_value;
				}
				else if(b[entry_index].type == "float")
				{
					b[entry_index].float_value = recursive_f(e2);
					return b[entry_index].float_value;
				}
				return recursive_f(e2);
			}
		}
		else if(rec_root.getText().equals("+="))
		{
			// assign to left node, left node must be ID
			e1 = (AST) rec_root.getFirstChild();	
			e2 = (AST) e1.getNextSibling();
			
			if( !(dict.containsKey(e1.getText()))) //this is number, make no sense to assign to number, so skip assing operator
			{			
				recursive_f(e2);
				return Float.parseFloat(e1.getText());
			}
			else // this is ID, could be assigned value
			{
				entry_index=(Integer)dict.get( e1.getText());
				
				if(b[entry_index].type == "int")
				{
					b[entry_index].int_value += (int) recursive_f(e2);
					return b[entry_index].int_value;
				}
				else if(b[entry_index].type == "float")
				{
					b[entry_index].float_value += recursive_f(e2);
					return b[entry_index].float_value;
				}
				return recursive_f(e2);
			}
		}
		else if(rec_root.getText().equals("-="))
		{
			// assign to left node, left node must be ID
			e1 = (AST) rec_root.getFirstChild();	
			e2 = (AST) e1.getNextSibling();
			
			if( !(dict.containsKey(e1.getText()))) //this is number, make no sense to assign to number, so skip assing operator
			{			
				recursive_f(e2);
				return Float.parseFloat(e1.getText());
			}
			else // this is ID, could be assigned value
			{
				entry_index=(Integer)dict.get( e1.getText());
				if(b[entry_index].type == "int")
				{
					b[entry_index].int_value -= (int) recursive_f(e2);
					return b[entry_index].int_value;
				}
				else if(b[entry_index].type == "float")
				{
					b[entry_index].float_value -= recursive_f(e2);
					return b[entry_index].float_value;
				}
				return recursive_f(e2);
			}
		}
		else if(rec_root.getText().equals("*="))
		{
			// assign to left node, left node must be ID
			e1 = (AST) rec_root.getFirstChild();	
			e2 = (AST) e1.getNextSibling();
			
			if( !(dict.containsKey(e1.getText()))) //this is number, make no sense to assign to number, so skip assing operator
			{			
				recursive_f(e2);
				return Float.parseFloat(e1.getText());
			}
			else // this is ID, could be assigned value
			{
				entry_index=(Integer)dict.get( e1.getText());
				
				if(b[entry_index].type == "int")
				{
					b[entry_index].int_value *= (int) recursive_f(e2);
					return b[entry_index].int_value;
				}
				else if(b[entry_index].type == "float")
				{
					b[entry_index].float_value *= recursive_f(e2);
					return b[entry_index].float_value;
				}
				return recursive_f(e2);
			}
		}
		else if(rec_root.getText().equals("/="))
		{
			// assign to left node, left node must be ID
			e1 = (AST) rec_root.getFirstChild();	
			e2 = (AST) e1.getNextSibling();
			
			if( !(dict.containsKey(e1.getText()))) //this is number, make no sense to assign to number, so skip assing operator
			{			
				recursive_f(e2);
				return Float.parseFloat(e1.getText());
			}
			else // this is ID, could be assigned value
			{
				entry_index=(Integer)dict.get( e1.getText());
				
				if(b[entry_index].type == "int")
				{
					b[entry_index].int_value /= (int) recursive_f(e2);
					return b[entry_index].int_value;
				}
				else if(b[entry_index].type == "float")
				{
					b[entry_index].float_value /= recursive_f(e2);
					return b[entry_index].float_value;
				}
				return recursive_f(e2);
			}
		}
		else if(rec_root.getText().equals("=="))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			if( recursive_f(e1) == recursive_f(e2))
				return 1;
			return -1;
		}
		
		else if(rec_root.getText().equals("!="))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			if( recursive_f(e1) != recursive_f(e2))
				return 1;
			return 0;
		}
		else if(rec_root.getText().equals("<"))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			if( recursive_f(e1) < recursive_f(e2))
				return 1;
			return -1;
		}
		else if(rec_root.getText().equals(">"))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			if( recursive_f(e1) > recursive_f(e2))
				return 1;
			return -1;
		}
		else if(rec_root.getText().equals("<="))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			if( recursive_f(e1) <= recursive_f(e2))
				return 1;
			return 0;
		}
		else if(rec_root.getText().equals(">="))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			if( recursive_f(e1) >= recursive_f(e2))
				return 1;
			return 0;
		}		
		else if(rec_root.getText().equals("&&"))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			int tempi1 = (int) recursive_f(e1);
			int tempi2 = (int) recursive_f(e2);
			if(tempi1>0 && tempi2>0)
				return 1;
			return 0;
		}
		else if(rec_root.getText().equals("||"))
		{
			e1 = (AST) rec_root.getFirstChild();
			e2 = (AST) e1.getNextSibling();
			int tempi1 = (int) recursive_f(e1);
			int tempi2 = (int) recursive_f(e2);
			if(tempi1>0 || tempi2>0)
				return 1;
			return 0;
		}
		else if(rec_root.getText().equals("break"))
		{
			if_break = 1;
			return 1;
		}
		else if(rec_root.getText().equals("continue"))
		{
			if_continue = 1;
			return 1;
		}
		else if(rec_root.getText().equals("++"))
		{
			// assign to left node, left node must be ID
			e1 = (AST) rec_root.getFirstChild();	
			entry_index=(Integer)dict.get( e1.getText());
				
			if(b[entry_index].type == "int")
			{
				return (b[entry_index].int_value++);
			}
			else if(b[entry_index].type == "float")
			{
				return (b[entry_index].float_value++);
			}
			return 0;
		}
		else if(rec_root.getText().equals("--"))
		{
			// assign to left node, left node must be ID
			e1 = (AST) rec_root.getFirstChild();	
			entry_index=(Integer)dict.get( e1.getText());
				
			if(b[entry_index].type == "int")
			{
				return (b[entry_index].int_value--);
			}
			else if(b[entry_index].type == "float")
			{
				return (b[entry_index].float_value--);
			}
			return 0;
		}		
		else
		{
			if( !(dict.containsKey(rec_root.getText()))) //this is number
			{			
				return Float.parseFloat(rec_root.getText());
			}
    		else //this is id
			{
				entry_index=(Integer)dict.get( rec_root.getText());
				
				if(b[entry_index].type == "int")
					return b[entry_index].int_value;
				else if(b[entry_index].type == "float")
					return b[entry_index].float_value;
				else
				{
					return 0;
					//not in our consideration now
				}		
			}				
		}
	}
	
}


program {int dummy; } 
: #(PROGRAM_AST {
                     kk.creatkml();
                     kk.sortCityCoords();
                }
     (dummy=stmt 
           {
            })+ 
            {kk.endkml();}
           ) 
;

stmt returns [int dummy]
{dummy=0;float ii=0.0f; float jj=0.0f;float kf=0.0f;double dd=0.0d;
int tempi1 = 0;
float tempf1 = 0.0f;
}
: #("dist" {
			b[count] = new GECLdatatypes();
			b[count].settype("distance");                } 
    ID     {b[count].setname(#ID.getText());             
    		dict.put(#ID.getText(),count);} 
    
    (DIST   {b[count].setdists( #DIST.getText()); })?
    { count++; }
    )  

| #("coord" {
			 b[count] = new GECLdatatypes();
			 b[count].settype("coord");                } 
    ID      {b[count].setname(#ID.getText());             
    		dict.put(#ID.getText(),count);} 
    {b[count].setcoord( ii,jj,"0ft");}
    (((MINUS  (m1:FLOAT   {ii=Float.parseFloat(#m1.getText()); ii=0-ii;}))|
    (f1:FLOAT   {ii=Float.parseFloat(#f1.getText());}))
    ( ( MINUS  (m2:FLOAT   {jj=Float.parseFloat(#m2.getText()); jj=0-jj;} ) ) |
    ( f2:FLOAT   {jj=Float.parseFloat(#f2.getText());}))
    DIST    {b[count].setcoord( ii,jj,#DIST.getText());})?
    {count++; }
    )

| #("AsigCity" {  } 
    as1:ID  {   if(!(dict.containsKey(#as1.getText()))) kk.reportError("not a valid city", #as1.getText());
                else dummy=(Integer)dict.get(#as1.getText());          
            }
    as2:StringConstant {}
    as3:StringConstant {
    WorldCoords tmp=kk.getCityCoords(#as2.getText(),#as3.getText());
    b[dummy].setCityCoords(tmp);}
    )    

| #("point" {
			 b[count] = new GECLdatatypes();
			 b[count].settype("point");
            } 
    s1:ID   {b[count].setname(#s1.getText());             
    		dict.put(#s1.getText(),count);} 
	s2:ID   {
            if(!(dict.containsKey(#s2.getText()))) kk.reportError("unrecognized", #s2.getText());
            else {
                  dummy=(Integer)dict.get(#s2.getText());
                  if(b[dummy].gettype()!="coord")
                    kk.reportError("not a coordinate", #s2.getText());
                  else 
                    b[count].cpy(b[dummy]);
            }
            count++;   
	        }
    )         

| #("place" {
			 b[count] = new GECLdatatypes();
			 b[count].settype("place");                } 
    s3:ID      {b[count].setname(#s3.getText());             
    		dict.put(#s3.getText(),count);} 
	s4:ID   {
			if(!(dict.containsKey(#s4.getText()))) kk.reportError("unrecognized", #s4.getText());
    		else {
    		      dummy=(Integer)dict.get(#s4.getText()); 
                  if(b[dummy].gettype()!="point")
                    kk.reportError("not a point", #s4.getText());
                  else 
                    b[count].cpy(b[dummy]);
    		     }         
	        }
	s5:StringConstant {
	                b[count].settitle(#s5.getText());
	               }
	(pla1:ID   {
			if(!(PerspTab.containsKey(#pla1.getText()))) kk.reportError("unrecognized", #pla1.getText());
    		else {
    		      dummy=(Integer)PerspTab.get(#pla1.getText()); 
                  b[count].setpersp(pr[dummy]);
    		     }         
	        })?            	             
	{ count++;  }        
    )             


| #("struct"{ 	b[count] = new GECLdatatypes();
			 	b[count].settype("struct");                	} 
    s6:ID   {	b[count].setname(#s6.getText());             
    			dict.put(#s6.getText(),count);				} 
    
    (((	"square" 	{b[count].setshape("square");}
    	f3:FLOAT   	{ii=Float.parseFloat(#f3.getText());}
    	f4:FLOAT   	{jj=Float.parseFloat(#f4.getText());}
		s7:ID		{
						if( !(dict.containsKey(#s7.getText()))) {
						kk.reportError("unrecognized: ",#s7.getText());
    					}else {
    		      		dummy=(Integer)dict.get(#s7.getText()); 
    		      		b[count].cpy(b[dummy]);
    		     		}
    		     	 	b[count].setsquare(ii,jj);           
	           		}
    )|("rectangle"  {b[count].setshape("rectangle");}
    	f5:FLOAT   	{ii=Float.parseFloat(#f5.getText());}
    	f6:FLOAT   	{jj=Float.parseFloat(#f6.getText());}
    	f7:FLOAT   	{kf=Float.parseFloat(#f7.getText());}
		s8:ID		{
						if( !(dict.containsKey(#s8.getText()))) {
						kk.reportError("unrecognized: ",#s8.getText());
    					}else {
    		      		dummy=(Integer)dict.get(#s8.getText()); 
    		      		b[count].cpy(b[dummy]);
    		     		}
    		     	 	b[count].setrect(ii,jj,kf);           
	        		}
    	
    )|("triangle" {b[count].setshape("triangle");})
   ))?
    { count++; }
   )
   

| #("overlay" 	{	b[count] = new GECLdatatypes();
					b[count].settype("overlay");                } 
    ovs1:ID   	{	b[count].setname(#ovs1.getText());             
    				dict.put(#ovs1.getText(),count);} 
	ovs2:ID   	{	if( !(dict.containsKey(#ovs2.getText()))) kk.reportError("unrecognized", #ovs2.getText());
    				else {
    		      		dummy=(Integer)dict.get(#ovs2.getText()); 
                        if(b[dummy].gettype()!="coord")
                            kk.reportError("not a coordinate", #ovs2.getText());
                        else 
                            tm[0]=b[dummy];
    		    	}         
	        	}
	ovs3:ID   	{	if( !(dict.containsKey(#ovs3.getText()))) kk.reportError("unrecognized", #ovs3.getText());
    				else {
    		      		dummy=(Integer)dict.get(#ovs3.getText()); 
                        if(b[dummy].gettype()!="coord")
                            kk.reportError("not a coordinate", #ovs3.getText());
                        else 
        		      		tm[1]=b[dummy];
    		     	}         
	        	}        
	ovs4:ID   	{	if( !(dict.containsKey(#ovs4.getText()))) kk.reportError("unrecognized", #ovs4.getText());
    				else {
    		      		dummy=(Integer)dict.get(#ovs4.getText()); 
                        if(b[dummy].gettype()!="coord")
                            kk.reportError("not a coordinate", #ovs3.getText());
                        else 
        		      		tm[2]=b[dummy];
    		     	}         
	        	}        
	ovs5:ID   	{	if( !(dict.containsKey(#ovs5.getText()))) kk.reportError("unrecognized", #ovs5.getText());
    				else {
    		      		dummy=(Integer)dict.get(#ovs5.getText()); 
                        if(b[dummy].gettype()!="coord")
                            kk.reportError("not a coordinate", #ovs5.getText());
                        else 
        		      		tm[3]=b[dummy];
    		     	}   
    		 		b[count].setoverlay(tm[0],tm[1],tm[2],tm[3]);    		           
	       	 	}        
	StringConstant {b[count].seturl(#StringConstant.getText());  }
	ovs6:FLOAT 	{   ii=Float.parseFloat(#ovs6.getText());
					b[count].setrotat(ii); 
					count++;  
				}             
 	)            
 	

| #("persp" { pr[countp] = new PerspectType(); } 
    psp1:ID      {pr[countp].setname(#psp1.getText());
                    PerspTab.put(#psp1.getText(),countp);} 
	psp2:ID   {
			if( !(dict.containsKey(#psp2.getText()))) kk.reportError("unrecognized", #psp2.getText());
    		else {
    		      dummy=(Integer)dict.get(#psp2.getText()); 
                  if(b[dummy].gettype()!="coord")
                    kk.reportError("not a coordinate", #psp2.getText());
                  else 
                    pr[countp].setcoord(b[dummy]);
    		     }         
	        }
	DIST    { 	    Distance Mydist = new Distance(#DIST.getText());
				pr[countp].setrange( Mydist.getDouble()/0.3048 );}
	psp3:FLOAT 	{   ii=Float.parseFloat(#psp3.getText());
					pr[countp].settilt(ii); 
				}         
	psp4:FLOAT 	{   jj=Float.parseFloat(#psp4.getText());
					pr[countp].setheading(jj); 
					countp++;  
				}         
    )              	
    

| #("line" {
			 b[count] = new GECLdatatypes();
			 b[count].settype("line");                } 
    ln1:ID      {b[count].setname(#ln1.getText());             
    			dict.put(#ln1.getText(),count);} 
	(ID   {
			if( !(dict.containsKey(#ID.getText()))) kk.reportError("unrecognized", #ID.getText());
            else {
    		      dummy=(Integer)dict.get(#ID.getText()); 
                  if(b[dummy].gettype()!="coord")
                    kk.reportError("not a coordinate", #ID.getText());
                  else 
    		        b[count].insert(b[dummy]);
    		     }         
	        })+     
	{ count++;  }        
    )                 
|  #("IS"
    type:ID     {   if( !(dict.containsKey(#type.getText()))) kk.reportError("unrecognized", #type.getText());
                    else {
                       dummy=(Integer)dict.get(#type.getText());
                       if(b[dummy].gettype()!="coord" && b[dummy].gettype()!="distance" && b[dummy].gettype()!="struct")
                            kk.reportError("invalid statement", "IS");
                       else if(((b[dummy].gettype()).toString()).equals("distance")) {
                            AST ast = type.getNextSibling();
                            String dstnc = ast.toString();
                            if(!(dict.containsKey(#ast.getText()))) kk.reportError("unrecognized", #ast.getText());
                            int dumm = (Integer)dict.get(#ast.getText());
                            if(b[dumm].gettype()!="distance") kk.reportError("invalid statement", "IS");
                            dstnc = b[dumm].getdists();
                            sep_dist = new Distance(dstnc);
                            int IntValue=0;
                            double tmpdouble=0.0;
                            float FloValue=0.0f;
                            if((ast.getNextSibling()).getNextSibling().getType()==DIST){
                                dstnc = ((ast.getNextSibling()).getNextSibling()).toString();                            	
                            	Distance new_dist = new Distance(dstnc);               
                            	tmpdouble= new_dist.getDouble(); // get meter
                            	FloValue = (float)tmpdouble;
                        
                        	}else if ((ast.getNextSibling()).getNextSibling().getType()==INTEGER){
                        		dstnc = ((ast.getNextSibling()).getNextSibling()).toString();
                            	IntValue = Integer.parseInt(dstnc);
                        
                        	}                      
                          //  dstnc = ((ast.getNextSibling()).getNextSibling()).toString();
                         //   int value = Integer.parseInt(dstnc);
                            if(((ast.getNextSibling()).toString()).equals("*")) {
                                dd = sep_dist.getDouble() * IntValue / 0.3048;
                                dstnc = Double.toString(dd) + "ft";
                                b[dummy].setdists(dstnc);
                            }
                            else {
                                dd = (sep_dist.getDouble() + FloValue) / 0.3048;
                                dstnc = Double.toString(dd) + "ft";
                                b[dummy].setdists(dstnc);
                            }
                        }else if(((b[dummy].gettype()).toString()).equals("coord"))
                        {
                           	float setdist=0.0f ;
                        	int setdir=0;
                        // get first ID	
                        	AST ast = type.getNextSibling();
                            String dstnc = ast.toString();
                            if(!(dict.containsKey(#ast.getText()))) kk.reportError("unrecognized", #ast.getText());
                            int dumm = (Integer)dict.get(#ast.getText());
                            if(!b[dumm].gettype().equals("coord")) kk.reportError("invalid statement", "IS");
                        	boolean IsPlus = false;
                        	if(((ast.getNextSibling()).toString()).equals("+")) {
                            IsPlus = true; 
                            }
                            else {
                            kk.reportError("invalid operator", "*");
                            }
                        // get first parameter suppose a dist ID or number   
                            AST ast1 = (ast.getNextSibling()).getNextSibling();
                            if(ast1.getType()==ID){
                        // get a dist ID    	
                            dstnc = ast1.toString();
                            if(!(dict.containsKey(#ast1.getText()))) kk.reportError("unrecognized", #ast1.getText());
                            int dummdist = (Integer)dict.get(#ast1.getText());
                            if(b[dummdist].gettype()!="distance") kk.reportError("invalid statement", "IS");
 							sep_dist = new Distance(b[dummdist].getdists());
 							setdist = (float)(sep_dist.getDouble()/ 0.3048);
 							
 							
 							}else if(ast1.getType()==DIST){
 						// get a dist number	
 							sep_dist = new Distance(#ast1.getText());
 							setdist = (float)(sep_dist.getDouble()/ 0.3048);
 							
 							}
						                    
                       // get second parameter suppose a direction         
                            AST ast2 = ast1.getNextSibling();
                            if(ast2.getType()==ID){
                        // get a dir ID    	
                            dstnc = ast2.toString();
                            if(!(dict.containsKey(#ast2.getText()))) kk.reportError("unrecognized", #ast2.getText());
                            int dummdir = (Integer)dict.get(#ast2.getText());
                            if(b[dummdir].gettype()!="int") kk.reportError("invalid statement", "IS");
 							setdir = b[dummdir].getIntValue();					
 							}else if(ast2.getType()==INTEGER){
 						// get a dir number	
 							setdir = Integer.parseInt(#ast2.getText());
 							}
                         b[dummy].navigate(b[dumm].getlon(),b[dumm].getlat(),setdist,setdir);   
                        }
                    }
                }
    )
    
|  #("printk" 
    ID     {
    		if( !(dict.containsKey(#ID.getText()))) kk.reportError("unrecognized: ",#ID.getText());
    		else {dummy=(Integer)dict.get(#ID.getText()); }         
    		if(b[dummy].gettype() == "distance") kk.printdis(b[dummy].getdists());
    		if(b[dummy].gettype() == "coord") kk.printcoord(b[dummy]);
    		if(b[dummy].gettype() == "point") kk.printPoint(b[dummy]);
    		if(b[dummy].gettype() == "place") kk.printPlace(b[dummy]);
    		if(b[dummy].gettype() == "struct"){ 
    		if(b[dummy].getshape() == "square") kk.printquadrangle(b[dummy]);
    		if(b[dummy].getshape() == "rectangle") kk.printquadrangle(b[dummy]);
    		if(b[dummy].getshape() == "triangle") /*{ tm.settriangle(b[dummy]);kk.printtriangle(b[dummy]);}*/;
    		}
    		if(b[dummy].gettype() == "overlay") kk.printOverlay(b[dummy]);
    		if(b[dummy].gettype() == "line") kk.printLine(b[dummy]);
			if(b[dummy].gettype() == "read3Dfile") b[dummy].read3Dfile.renderKML(kk.outp);
    	   } 
    )

|  #("int" {
			 b[count] = new GECLdatatypes();
			 b[count].settype("int");                } 
    ID      {b[count].setname(#ID.getText());             
    		dict.put(#ID.getText(),count);} 
			
	i1:INTEGER {
					tempi1=Integer.parseInt(#i1.getText());
	                b[count].setIntValue(tempi1);      
	                 count++;  
	               }      
    )   

|  #("float" {
			 b[count] = new GECLdatatypes();
			 b[count].settype("float");                } 
    ID      {b[count].setname(#ID.getText());             
    		dict.put(#ID.getText(),count);} 
			
	f21:FLOAT {
					tempf1=Float.parseFloat(#f21.getText());
	                b[count].setFloatValue(tempf1);      
	                 count++;  
	               }      
    )   	
| #("readfile" {
			 b[count] = new GECLdatatypes();
			 b[count].settype("read3Dfile");
			 } 
    ID      {b[count].setname(#ID.getText());             
    		dict.put(#ID.getText(),count);
			} 
		
	( scale1:FLOAT {
		b[count].read3Dfile.scale = Float.parseFloat(#scale1.getText());
	}
	| ID {
		tempi1 = (Integer)dict.get(#ID.getText());
		b[count].read3Dfile.scale = b[tempi1].getFloatValue();
		}
	)
	
	( translate_x1:FLOAT {
		b[count].read3Dfile.translate_x = Float.parseFloat(#translate_x1.getText());
	}
	| ID {
		tempi1 = (Integer)dict.get(#ID.getText());
		b[count].read3Dfile.translate_x = b[tempi1].getFloatValue();
		}
	)
	
	( translate_y1:FLOAT {
		b[count].read3Dfile.translate_y = Float.parseFloat(#translate_y1.getText());
	}
	| ID {
		tempi1 = (Integer)dict.get(#ID.getText());
		b[count].read3Dfile.translate_y = b[tempi1].getFloatValue();
		}
	)
	
	(translate_z1:FLOAT {
		b[count].read3Dfile.translate_z = Float.parseFloat(#translate_z1.getText());
	}
	| ID {
		tempi1 = (Integer)dict.get(#ID.getText());
		b[count].read3Dfile.translate_z = b[tempi1].getFloatValue();
		}
	)
	
	st21:StringConstant {
					b[count].read3Dfile.openfile(#st21.getText());
	                 count++;  
	               }
    )
| statement
;

statement
{
	float exprReturn = 0; 
	float ii = 0.0f;
	float jj = 0.0f;
	float kf = 0.0f;
	double dd = 0;
	AST e1, e2, e3, as2, as3;
	int dummy = 0;
}
        :       s:SEMI                
		|       #(NStatementExpr
                exprReturn = expr 
                )                 // Expressions
// Iteration statements:
        |       #( w:"while" {if_break = 0; if_skip_rcurly = 0;}
					exprReturn = exp1:expr
					(
						{ exprReturn > 0 }? block1:block_statement
							{		
								if(if_break == 1)
									if_break = 0;
								else
								{
								
									while( recursive_f(exp1) > 0)
									{
										e1 = (AST) w.getFirstChild(); //expr
										e1 = (AST) e1.getNextSibling(); // {
										e1 = (AST) e1.getNextSibling(); // first statement now
										while(true) 
										{
											if(e1.getText().length()>0)
											{
												if(e1.getText().equals("break"))
												{
													if_break = 1;
													break;
												}
												else if(e1.getText().equals("if"))
												{
													e2 = (AST) e1.getFirstChild(); // expr => ==, ||, &&, != .......
													e3 = (AST) e2.getNextSibling(); // {
													e3 = (AST) e3.getNextSibling(); // block statement
													int tempi1 = (int) recursive_f(e2);
													if(tempi1 > 0) // execut this block
													{
														while(!e3.getText().equals("}"))
														{
															recursive_f(e3);
															e3 = (AST) e3.getNextSibling();
														}
													}
													else
													{
														//skip first block
														while(!e3.getText().equals("}"))
														{
															e3 = (AST) e3.getNextSibling();
														}
														e3 = (AST) e3.getNextSibling(); //check "else"
    													if(e3 != null) //this is else
														{
															e3 = (AST) e3.getNextSibling(); // "{"
															e3 = (AST) e3.getNextSibling(); // statement block
															while(!e3.getText().equals("}"))
															{
																recursive_f(e3);
																e3 = (AST) e3.getNextSibling();
															}
														}
													}
													if(if_break == 1 || if_continue == 1)
													{
														break;
													}
												}
												else if(e1.getText().equals("continue") || e1.getText().equals("}"))
												{
													break;
												}
												else if(e1.getText().equals("dist"))
												{
													//dist
													b[count] = new GECLdatatypes();
													b[count].settype("distance");
													//ID	
													e2 = e1.getFirstChild();
													b[count].setname(e1.getText());             
													dict.put(e2.getText(),count);
													//(DIST )?
													e2 = e2.getNextSibling();
													if(e2!=null)
													b[count].setdists( e2.getText());
													count++;
												}
												else if(e1.getText().equals("coord"))
												{
													//coord
													b[count] = new GECLdatatypes();
													b[count].settype("coord");
													//ID
													e2 = e1.getFirstChild();
													b[count].setname(e2.getText());             
													dict.put(e2.getText(),count);
													b[count].setcoord( ii,jj,"0ft");
												
													//FLOAT | (MINUS FLOAT)
													e2 = e2.getNextSibling();
												
													if(e2 == null) //finished, there is no assign value here
													{
													}
													else // assign value
													{
														if(e2.getText().equals("-")) //MINUS FLOAT case
														{
															e2 = e2.getNextSibling();
															ii=Float.parseFloat(e2.getText()); 
															ii=0-ii;
														}
														else //FLOAT
														{
															e2 = e2.getNextSibling();
															ii=Float.parseFloat(e2.getText());
														}
													
														//FLOAT | (MINUS FLOAT)
														e2 = e2.getNextSibling();
														if(e2.getText().equals("-")) //MINUS FLOAT case
														{
															e2 = e2.getNextSibling();
															jj=Float.parseFloat(e2.getText()); 
															jj=0-jj;
														}
														else //FLOAT
														{
															e2 = e2.getNextSibling();
															jj=Float.parseFloat(e2.getText());
														}
														//DIST
														e2 = e2.getNextSibling();
														b[count].setcoord( ii,jj, e2.getText());
													}
													count++;
												}
												else if(e1.getText().equals("AsigCity"))
												{
													//as1:ID
													e2 = e1.getFirstChild();
													if(!(dict.containsKey(e2.getText())))
													{													
														kk.reportError("not a valid city", e2.getText());
													}
													else
													{													
														dummy=(Integer)dict.get(e2.getText());
														
														//as2:StringConstant
														e2 = e2.getNextSibling();
														as2 = e2;
													
														//as3:StringConstant
														e2 = e2.getNextSibling();
														as3 = e2;
													
														WorldCoords tmp=kk.getCityCoords(as2.getText(), as3.getText());
														b[dummy].setCityCoords(tmp);
													}
												}
												else if(e1.getText().equals("point"))
												{
													//point
													b[count] = new GECLdatatypes();
													b[count].settype("point");
													//s1:ID	
													e2 = e1.getFirstChild();
													b[count].setname(e2.getText());             
													dict.put(e2.getText(),count);
													//s2:ID
													e2 = e2.getNextSibling();
													if(!(dict.containsKey(e2.getText()))) 
														kk.reportError("unrecognized", e2.getText());
													else 
													{
														dummy=(Integer)dict.get(e2.getText());
														if(b[dummy].gettype()!="coord")
															kk.reportError("not a coordinate", e2.getText());
														else 
															b[count].cpy(b[dummy]);
													}
													count++;
												}
												else if(e1.getText().equals("place"))
												{
													//place
													b[count] = new GECLdatatypes();
													b[count].settype("place");
													//s3:ID
													e2 = e1.getFirstChild();
													b[count].setname(e2.getText());             
													dict.put(e2.getText(),count);
													//s4:ID
													e2 = e2.getNextSibling();
													if(!(dict.containsKey(e2.getText()))) 
														kk.reportError("unrecognized", e2.getText());
													else 
													{
														dummy=(Integer)dict.get(e2.getText()); 
														if(b[dummy].gettype()!="point")
															kk.reportError("not a point", e2.getText());
														else 
															b[count].cpy(b[dummy]);
													}
													//s5:StringConstant
													e2 = e2.getNextSibling();
													b[count].settitle(e2.getText());
													// (pla1:ID)?
													e2 = e2.getNextSibling();
													if(e2 == null)
													{
													}
													else
													{
														if(!(PerspTab.containsKey(e2.getText()))) 
															kk.reportError("unrecognized", e2.getText());
														else 
														{
															dummy=(Integer)PerspTab.get(e2.getText()); 
															b[count].setpersp(pr[dummy]);
														}	
													}
													count++;
												}
												else if(e1.getText().equals("struct"))
												{
													//struct
													b[count] = new GECLdatatypes();
													b[count].settype("struct");
													//s6:ID
													e2 = e1.getFirstChild();
													b[count].setname(e2.getText());             
													dict.put(e2.getText(),count);
													//( shape ID )?
													e2 = e2.getNextSibling();
													if(e2==null)
													{
													}
													else
													{
														if(e2.getText().equals("square"))
														{
															b[count].setshape("square");
															//f3:FLOAT
															e2 = e2.getNextSibling();
															ii=Float.parseFloat(e2.getText());
															//f4:FLOAT
															e2 = e2.getNextSibling();
															jj=Float.parseFloat(e2.getText());
															//s7:ID
															e2 = e2.getNextSibling();
															if(!(dict.containsKey(e2.getText()))) 
															{
																kk.reportError("unrecognized: " , e2.getText());
															}
															else
															{
																dummy=(Integer)dict.get(e2.getText()); 
																b[count].cpy(b[dummy]);
															}
															b[count].setsquare(ii,jj);			
														}
														else if(e2.getText().equals("rectangle"))
														{
															b[count].setshape("rectangle");
															//f5:FLOAT
															e2 = e2.getNextSibling();
															ii=Float.parseFloat(e2.getText());
															//f6:FLOAT
															e2 = e2.getNextSibling();
															jj=Float.parseFloat(e2.getText());
															//f7:FLOAT
															e2 = e2.getNextSibling();
															kf=Float.parseFloat(e2.getText());
															//s8:ID
															e2 = e2.getNextSibling();
															if( !(dict.containsKey(e2.getText()))) 
															{
																kk.reportError("unrecognized: ", e2.getText());
															}
															else 
															{
																dummy=(Integer)dict.get(e2.getText()); 
																b[count].cpy(b[dummy]);
															}
															b[count].setrect(ii,jj,kf);
														}
														else if(e2.getText().equals("triangle"))
														{
															b[count].setshape("triangle");
														}
														count++;
													}
												}
												else if(e1.getText().equals("overlay"))
												{
	//overlay
	b[count] = new GECLdatatypes();
	b[count].settype("overlay");
	
	//ovs1:ID
	e2 = e1.getFirstChild();
	b[count].setname(e2.getText());             
    dict.put(e2.getText(),count);
	
	//ovs2:ID
	e2 = e2.getNextSibling();
	if( !(dict.containsKey(e2.getText()))) 
		kk.reportError("unrecognized", e2.getText());
    else 
	{
    	dummy=(Integer)dict.get(e2.getText()); 
        if(b[dummy].gettype()!="coord")
            kk.reportError("not a coordinate", e2.getText());
        else 
            tm[0]=b[dummy];
    }
	
	//ovs3:ID
	e2 = e2.getNextSibling();
	if( !(dict.containsKey(e2.getText()))) 
		kk.reportError("unrecognized", e2.getText());
    else 
	{
    	dummy=(Integer)dict.get(e2.getText()); 
        if(b[dummy].gettype()!="coord")
            kk.reportError("not a coordinate", e2.getText());
        else 
       		tm[1]=b[dummy];
    }
	
	//ovs4:ID
	e2 = e2.getNextSibling();
	if( !(dict.containsKey(e2.getText()))) 
		kk.reportError("unrecognized", e2.getText());
    else 
	{
    	dummy=(Integer)dict.get(e2.getText()); 
        if(b[dummy].gettype()!="coord")
            kk.reportError("not a coordinate", e2.getText());
        else 
       		tm[2]=b[dummy];
    }
	
	//ovs5:ID
	e2 = e2.getNextSibling();
	if( !(dict.containsKey(e2.getText()))) 
		kk.reportError("unrecognized", e2.getText());
	else 
	{
		dummy=(Integer)dict.get(e2.getText()); 
        if(b[dummy].gettype()!="coord")
            kk.reportError("not a coordinate", e2.getText());
        else 
        	tm[3]=b[dummy];
    }   
    b[count].setoverlay(tm[0],tm[1],tm[2],tm[3]);
					
	//StringConstant
	e2 = e2.getNextSibling();
	b[count].seturl(e2.getText());
	
	//ovs6:FLOAT
	e2 = e2.getNextSibling();
	ii=Float.parseFloat(e2.getText());
	b[count].setrotat(ii); 
	count++;
}
else if(e1.getText().equals("persp"))
{
	//persp
	pr[countp] = new PerspectType(); 
	
	//psp1:ID
	e2 = e1.getFirstChild();
	pr[countp].setname(e2.getText());
    PerspTab.put(e2.getText(),countp);
	
	//psp2:ID
	e2 = e2.getNextSibling();
	if( !(dict.containsKey(e2.getText()))) 
		kk.reportError("unrecognized", e2.getText());
    else 
	{
        dummy=(Integer)dict.get(e2.getText()); 
        if(b[dummy].gettype()!="coord")
            kk.reportError("not a coordinate", e2.getText());
        else 
            pr[countp].setcoord(b[dummy]);
    }
	
	//DIST
	e2 = e2.getNextSibling();
	Distance Mydist = new Distance(e2.getText());
	pr[countp].setrange( Mydist.getDouble()/0.3048 );
	
	//psp3:FLOAT
	e2 = e2.getNextSibling();
	ii=Float.parseFloat(e2.getText());
	pr[countp].settilt(ii);
	
	//psp4:FLOAT
	e2 = e2.getNextSibling();
	jj=Float.parseFloat(e2.getText());
	pr[countp].setheading(jj); 
	countp++;
}
else if(e1.getText().equals("line"))
{
	//line
	b[count] = new GECLdatatypes();
	b[count].settype("line");
	
	//ln1:ID
	e2 = e1.getFirstChild();
	b[count].setname(e2.getText());             
    dict.put(e2.getText(),count);
	
	// ( ID )+
	e2 = e2.getNextSibling();
	while(e2 != null)
	{
		if( !(dict.containsKey(e2.getText())))
			kk.reportError("unrecognized", e2.getText());
        else 
		{
    	    dummy=(Integer)dict.get(e2.getText()); 
            if(b[dummy].gettype()!="coord")
                kk.reportError("not a coordinate", e2.getText());
            else 
				b[count].insert(b[dummy]);
		}
		e2 = e2.getNextSibling();
	}
	count++;
	
}

else if(e1.getText().equals("IS"))
{
	//IS, type:ID
	e2 = e1.getFirstChild();
	
	if( !(dict.containsKey(e2.getText()))) 
		kk.reportError("unrecognized", e2.getText());
                    else {
                       dummy=(Integer)dict.get(e2.getText());
                       if(b[dummy].gettype()!="coord" && b[dummy].gettype()!="distance" && b[dummy].gettype()!="struct")
                            kk.reportError("invalid statement", "IS");
                       else if(((b[dummy].gettype()).toString()).equals("distance")) {
                            AST ast = e2.getNextSibling();
                            String dstnc = ast.toString();
                            if(!(dict.containsKey(ast.getText()))) kk.reportError("unrecognized", ast.getText());
                            int dumm = (Integer)dict.get(ast.getText());
                            if(b[dumm].gettype()!="distance") kk.reportError("invalid statement", "IS");
                            dstnc = b[dumm].getdists();
                            sep_dist = new Distance(dstnc);
                            int IntValue=0;
                            double tmpdouble=0.0;
                            float FloValue=0.0f;
                            if((ast.getNextSibling()).getNextSibling().getType()==DIST){
                                dstnc = ((ast.getNextSibling()).getNextSibling()).toString();                            	
                            	Distance new_dist = new Distance(dstnc);               
                            	tmpdouble= new_dist.getDouble(); // get meter
                            	FloValue = (float)tmpdouble;
                        
                        	}else if ((ast.getNextSibling()).getNextSibling().getType()==INTEGER){
                        		dstnc = ((ast.getNextSibling()).getNextSibling()).toString();
                            	IntValue = Integer.parseInt(dstnc);
                        
                        	}                      
                          //  dstnc = ((ast.getNextSibling()).getNextSibling()).toString();
                         //   int value = Integer.parseInt(dstnc);
                            if(((ast.getNextSibling()).toString()).equals("*")) {
                                dd = sep_dist.getDouble() * IntValue / 0.3048;
                                dstnc = Double.toString(dd) + "ft";
                                b[dummy].setdists(dstnc);
                            }
                            else {
                                dd = (sep_dist.getDouble() + FloValue) / 0.3048;
                                dstnc = Double.toString(dd) + "ft";
                                b[dummy].setdists(dstnc);
                            }
                        }else if(((b[dummy].gettype()).toString()).equals("coord"))
                        {
                          	float setdist=0.0f ;
                        	int setdir=0;
                        // get first ID	
                        	AST ast = e2.getNextSibling();
                            String dstnc = ast.toString();
                            if(!(dict.containsKey(#ast.getText()))) kk.reportError("unrecognized", #ast.getText());
                            int dumm = (Integer)dict.get(#ast.getText());
                            if(!b[dumm].gettype().equals("coord")) kk.reportError("invalid statement", "IS");
                        	boolean IsPlus = false;
                        	if(((ast.getNextSibling()).toString()).equals("+")) {
                            IsPlus = true; 
                            }
                            else {
                            kk.reportError("invalid operator", "*");
                            }
                        // get first parameter suppose a dist ID or number   
                            AST ast1 = (ast.getNextSibling()).getNextSibling();
                            if(ast1.getType()==ID){
                        // get a dist ID    	
                            dstnc = ast1.toString();
                            if(!(dict.containsKey(#ast1.getText()))) kk.reportError("unrecognized", #ast1.getText());
                            int dummdist = (Integer)dict.get(#ast1.getText());
                            if(b[dummdist].gettype()!="distance") kk.reportError("invalid statement", "IS");
 							sep_dist = new Distance(b[dummdist].getdists());
 							setdist = (float)(sep_dist.getDouble()/ 0.3048);
 							
 							
 							}else if(ast1.getType()==DIST){
 						// get a dist number	
 							sep_dist = new Distance(#ast1.getText());
 							setdist = (float)(sep_dist.getDouble()/ 0.3048);
 							
 							}
						                    
                       // get second parameter suppose a direction         
                            AST ast2 = ast1.getNextSibling();
                            if(ast2.getType()==ID){
                        // get a dir ID    	
                            dstnc = ast2.toString();
                            if(!(dict.containsKey(#ast2.getText()))) kk.reportError("unrecognized", #ast2.getText());
                            int dummdir = (Integer)dict.get(#ast2.getText());
                            if(b[dummdir].gettype()!="int") kk.reportError("invalid statement", "IS");
 							setdir = b[dummdir].getIntValue();					
 							}else if(ast2.getType()==INTEGER){
 						// get a dir number	
 							setdir = Integer.parseInt(#ast2.getText());
 							}
                         b[dummy].navigate(b[dumm].getlon(),b[dumm].getlat(),setdist,setdir);   
                        }}	
}
else if(e1.getText().equals("printk"))
{
	//printk, ID
	
	e2 = e1.getFirstChild();
	if( !(dict.containsKey(e2.getText()))) 
		kk.reportError( "unrecognized: ", e2.getText() );
    else { 
		dummy = (Integer)dict.get(e2.getText()); 
	}
	
    if(b[dummy].gettype() == "distance") 
		kk.printdis(b[dummy].getdists());
    if(b[dummy].gettype() == "coord") 
		kk.printcoord(b[dummy]);
    if(b[dummy].gettype() == "point") 
		kk.printPoint(b[dummy]);
    if(b[dummy].gettype() == "place") 
		kk.printPlace(b[dummy]);
    if(b[dummy].gettype() == "struct")
	{ 
    	if(b[dummy].getshape() == "square") 
			kk.printquadrangle(b[dummy]);
    	if(b[dummy].getshape() == "rectangle") 
			kk.printquadrangle(b[dummy]);
    	if(b[dummy].getshape() == "triangle") ;
    }
    if(b[dummy].gettype() == "overlay") 
		kk.printOverlay(b[dummy]);
    if(b[dummy].gettype() == "line") 
		kk.printLine(b[dummy]);
	if(b[dummy].gettype() == "read3Dfile") 
		b[dummy].read3Dfile.renderKML(kk.outp);
	
}
else if(e1.getText().equals("readfile"))
{
	//readfile
	b[count] = new GECLdatatypes();
	b[count].settype("read3Dfile");
	
	//ovs1:ID
	e2 = e1.getFirstChild();
	b[count].setname(e2.getText());             
    dict.put(e2.getText(),count);
	
	//scale1:FLOAT
	e2 = e2.getNextSibling();
	if( !(dict.containsKey(e2.getText()))) 
		b[count].read3Dfile.scale = Float.parseFloat(e2.getText());
    else { 
		dummy = (Integer)dict.get(e2.getText()); 
		b[count].read3Dfile.scale = b[dummy].getFloatValue();
	} 
	
	//translate_x1:FLOAT
	e2 = e2.getNextSibling();
	if( !(dict.containsKey(e2.getText()))) 
		b[count].read3Dfile.translate_x = Float.parseFloat(e2.getText());
    else { 
		dummy = (Integer)dict.get(e2.getText()); 
		b[count].read3Dfile.translate_x = b[dummy].getFloatValue();
	}
	
	//translate_y1:FLOAT
	e2 = e2.getNextSibling();
	if( !(dict.containsKey(e2.getText()))) 
		b[count].read3Dfile.translate_y = Float.parseFloat(e2.getText());
    else { 
		dummy = (Integer)dict.get(e2.getText()); 
		b[count].read3Dfile.translate_y = b[dummy].getFloatValue();
	}
	
	//translate_z1:FLOAT
	e2 = e2.getNextSibling();
	if( !(dict.containsKey(e2.getText()))) 
		b[count].read3Dfile.translate_z = Float.parseFloat(e2.getText());
    else { 
		dummy = (Integer)dict.get(e2.getText()); 
		b[count].read3Dfile.translate_z = b[dummy].getFloatValue();
	}
	
	//filename
	e2 = e2.getNextSibling();
	b[count].read3Dfile.openfile(e2.getText());
	count++;  
    
}







												
											}											
											recursive_f(e1.getFirstChild());
											e1 = (AST) e1.getNextSibling();
										}
										if(if_break == 1)
										{
											if_break = 0;
											break;
										}
										if_continue = 0;
									}
								}
							}
						| skip_block_statement
					)
				)             

        |       #( d:"do" { if_break = 0; if_skip_rcurly = 0;}
                block2:block_statement  
				"while"    
                exprReturn = exp2:expr 
							{
    							if(if_break == 1)
								{
									if_break = 0;
								}
								else
								{
									while( exprReturn > 0)
									{
										e1 = (AST) d.getFirstChild(); //{
										e1 = (AST) e1.getNextSibling(); // first statement now
										while(true) 
										{
											if(e1.getText().length()>0)
											{
												if(e1.getText().equals("break"))
												{
													if_break = 1;
													break;
												}
												else if(e1.getText().equals("continue") || e1.getText().equals("}"))
												{
													break;
												}												
											}											
											recursive_f(e1.getFirstChild());
											e1 = (AST) e1.getNextSibling();
										}
										if(if_break == 1)
										{
											if_break = 0;
											break;
										}
										exprReturn = recursive_f(exp2);
										
									}
								}
								
							}
							
                        
                )
// Jump statements:
        |       "continue" {if_skip_rcurly = 1 ;} 
				//( skip_statement )+ RCURLY
        |       "break"               { if_break = 1 ; if_skip_rcurly = 1; }
				//( skip_statement )+ RCURLY
        |       #( r:"return"           
                ( exprReturn = expr )? 
                )
// Selection statements:
        |       #( "if" exprReturn = expr  
					( 
						{ exprReturn > 0 }? block_statement ( "else" skip_block_statement )?
						|   
							 skip_block_statement ("else" block_statement )?
							
					)
				)
		;
		
skip_statement
        :       s:SEMI 
		|       #( NStatementExpr . )
// Iteration statements:
        |       #( "while" . skip_block_statement )
        |       #( "do" skip_block_statement .)
// Jump statements:
        |       "continue"
        |       "break"
        |       #( "return" ( . )? )
// Selection statements:
        |       #( "if" . skip_block_statement ("else" skip_block_statement )?
				)
		;

skip_block_statement
		:	skip_statement
		|	lc:LCURLY					
			( skip_statement )+
			rc:RCURLY
		;
		

		
block_statement
{
	int tempi;
}
		:	tempi = stmt
		|	lc:LCURLY					{ tabs++; }
			( 
			{if_skip_rcurly == 0}? tempi = stmt
			| tempi = stmt
			)+ 
			rc:RCURLY					{ if_skip_rcurly = 0 ;tabs--; }
		;


/*
block_statement
		:	statement
		|	lc:LCURLY					{ tabs++; }
			( 
			{if_skip_rcurly == 0}? statement 
			| skip_statement
			)+ 
			rc:RCURLY					{ if_skip_rcurly = 0 ;tabs--; }
			//({if_skip_rcurly == 0}? rc:RCURLY)
			//{if_skip_rcurly = 0; tabs--; }
		;
		*/
		
expr returns [float binReturn]
{
int exprReturn = 0;
binReturn = 0; 
}
        :      binReturn = binaryExpr
        |       emptyExpr
        ;

emptyExpr
        :   NEmptyExpression
        ;
		
binaryOperator
        :       ASSIGN
        |       DIV_ASSIGN
        |       PLUS_ASSIGN
        |       MINUS_ASSIGN
        |       STAR_ASSIGN
        |       MOD_ASSIGN
        |       RSHIFT_ASSIGN
        |       LSHIFT_ASSIGN
        |       BAND_ASSIGN
        |       BOR_ASSIGN
        |       BXOR_ASSIGN
        |       LOR
        |       LAND
        |       BOR
        |       BXOR
        |       BAND
        |       EQUAL
        |       NOT_EQUAL
        |       LT
        |       LTE
        |       GT
        |       GTE
        |       LSHIFT
        |       RSHIFT
        |       PLUS
        |       MINUS
        |       STAR
        |       DIV
        |       MOD
        |       NCommaExpr
		|		INC
		|		DEC
        ;
		

binaryExpr returns [float binReturn]
{
	binReturn = 0;
	int entry_index1 = 0;
}
        :   		
			b:binaryOperator
            // no rules allowed as roots, so here I manually get 
            // the first and second children of the binary operator
            // and then print them out in the right order
            {   
				binReturn = recursive_f(b);				
            } 							                                  
        ;





